from __future__ import division
from builtins import zip
from past.utils import old_div
import numpy as np
from outliers import smirnov_grubbs as grubbs
import pandas as pd

from normalisation import constants
from normalisation import data_model as dm


def get_outlier_samples(internal_std_df):
    """
    This function returns outlier samples after performing grubb's test
    :param internal_std_df: Dataframe having intensity data of internal standard
    :return: Dataframe having samples which are outliers
    """
    outlier_data = grubbs.max_test_outliers(
        internal_std_df[constants.MQ_INTENSITY],
        alpha=0.05)

    rejected_samples = internal_std_df[
        internal_std_df[constants.MQ_INTENSITY].isin(
            outlier_data)]

    for sample in list(rejected_samples['Area']):
        internal_std_df = internal_std_df[internal_std_df['Area'] != sample]

    return rejected_samples, internal_std_df


def get_max_intensity_accepted_sample(accpted_samples_df):
    """
    This function returns dataframe having maximum intensity sample among
    acceptable samples after performing grubb's test
    :param internal_std_df: Dataframe having intensity data of internal standard
    :return: Dataframe having sample with maximum intensity among accepted
    samples
    """

    accepted_intensity = grubbs.test(accpted_samples_df[constants.MQ_INTENSITY],
                                     alpha=0.05)

    max_intensity_sample = accpted_samples_df.loc[accpted_samples_df[
                                                   constants.MQ_INTENSITY] == max(
        accepted_intensity)]
    return max_intensity_sample


def calculate_normalising_factor(raw_df, metab_name, sample_name=None):
    """
    Calculate normalisation factor for each sample by dividing the
    maximum intensity of internal standard across all the samples

    :param raw_df: raw dataframe having intensities of all metabolites
    :param metab_name: name of metabolite selected as internal standard
    :param sample_name: sample selected as maximum intensity sample
    :return: list containing normalisation factor for each sample.
    """
    internal_std_df = raw_df[raw_df[constants.PARENT] == metab_name]
    outlier_samples, accepted_samples_df = get_outlier_samples(internal_std_df)
    max_intensity_def_sample = get_max_intensity_accepted_sample(accepted_samples_df)
    intensity_is = internal_std_df[constants.MQ_INTENSITY]
    if sample_name:
        max_intensity_sample = internal_std_df.loc[internal_std_df[constants.MQ_SAMPLE_NAME] == sample_name]
        max_intensity = max_intensity_sample[constants.MQ_INTENSITY]
    else:
        max_intensity = max_intensity_def_sample[constants.MQ_INTENSITY].unique()
    try:
        norm_factor = [float(x) / float(max_intensity) for x in intensity_is]
    except:
        norm_factor = [float(x) for x in intensity_is]
    return norm_factor, outlier_samples, max_intensity_def_sample


def normalised_data(raw_df, metab_name, norm_factor):
    """
    Calculates normalised intensity data for the metabolite passed.

    :param raw_df: raw dataframe having intensities of all metabolites
    :param metab_name: name of metabolite for with normalised intensity is
                       calculated.
    :param norm_factor: list having normalisation factor for each sample
    :return: normalised intensity for the metabolite passed in function
    """
    metab_df = raw_df[raw_df[constants.PARENT] == metab_name]
    intensity_mb = metab_df[constants.MQ_INTENSITY]
    norm_intensity = []
    for i, nf in zip(intensity_mb, norm_factor):
        norm_intensity.append(old_div(i, nf))
    norm_intensity_list = np.array(norm_intensity)
    norm_intensity_list[norm_intensity_list == np.inf] = 0
    return norm_intensity_list


def normalized_intensity_df(raw_df, metadata_mq, internal_std_name,
                            sample_name):
    """
    Dataframe with normalised intensity of all the metabolites present in the
    samples
    :param raw_df: raw dataframe having intensities of all metabolites
    :param metadata_mq: dataframe having mapping of unlabeled fragment with
                        metabolite fragment
    :param internal_std_name: name of internal standard
    :param sample_name: sample selected as maximum intensity sample
    :return: raw dataframe with normalised intensities values of all
             metabolites with respect to internal standard
    """
    raw_df = dm.merge_mq_metadata(raw_df, metadata_mq)
    raw_df = dm.compute_sum_by_fragments(raw_df)
    norm_factor, outlier_samples, max_intensity_def_sample = calculate_normalising_factor(
        raw_df, internal_std_name,
        sample_name)
    metab_list = raw_df[constants.PARENT].unique()
    raw_norm_df = pd.DataFrame()
    for metabolite in metab_list:
        norm_intensity = normalised_data(raw_df, metabolite, norm_factor)
        metab_df = raw_df[raw_df[constants.PARENT] == metabolite]
        norm_fac_df = pd.concat([metab_df.reset_index(drop=True), pd.DataFrame(
            norm_intensity)], axis=1)
        norm_fac_df = norm_fac_df.rename(columns={constants.MQ_INTENSITY:
                                                      constants.RAW_AREA})
        norm_fac_df = norm_fac_df.rename(columns={0: constants.MQ_INTENSITY})
        raw_norm_df = raw_norm_df.append(norm_fac_df, ignore_index=True)
    raw_norm_df = raw_norm_df.fillna(0)
    return raw_norm_df


def get_normalized_biosample_data(raw_df, metadata_std,
                                  metadata_mq, internal_std_name, sample_name):
    """
    This function returns standard and biosample dataframe having normalised
    intensity of all the metabolites present in the samples.
    :param raw_df: raw dataframe having intensities of all metabolites
    :param metadata_std: standard metadata dataframe containing concentration
                        mapping of the standard samples
    :param metadata_mq: dataframe having mapping of unlabeled fragment with
                        metabolite fragment
    :param internal_std_name: name of internal standard
    :param sample_name: sample selected as maximum intensity sample
    :return: standard and biosamples dataframe with normalised intensity values.
    """
    norm_df = normalized_intensity_df(raw_df, metadata_mq, internal_std_name,
                                      sample_name)
    standard_df, biosamples_df = dm.split_std_bio(norm_df, metadata_std)
    standard_df = dm.merge_std_df_metadata(metadata_std, standard_df)
    columns = [constants.MQ_INTENSITY, constants.MQ_CONCENTRATION]
    biosamples_df[columns] = biosamples_df[columns].applymap \
        (lambda x: np.round(x, decimals=constants.PRESICION))
    return standard_df, biosamples_df


def get_outliers_data(raw_df, metadata_mq):
    """
    This function returns dictionary having information of default maximum
    intensity sample and outlier samples for all the metabolites present.
    :param raw_df: raw dataframe having intensities of all metabolites
    :param metadata_mq:dataframe having mapping of unlabeled fragment with
                        metabolite fragment
    :return:dictionary having information of default maximum
    intensity sample and outlier samples for all the metabolites present.
    """
    raw_df = dm.merge_mq_metadata(raw_df, metadata_mq)
    raw_df = dm.compute_sum_by_fragments(raw_df)
    metab_list = raw_df[constants.PARENT].unique()
    outlier_test = []
    for metabolite in metab_list:
        norm_factor, outlier_samples, max_intensity_def_sample = calculate_normalising_factor(
            raw_df, metabolite, sample_name=None)
        outlier_samples = outlier_samples[[constants.MQ_SAMPLE_NAME,
                                           constants.MQ_INTENSITY]]
        max_intensity_def_sample = max_intensity_def_sample[[constants.MQ_SAMPLE_NAME,
                                                             constants.MQ_INTENSITY]]
        outlier_test.append({'name': metabolite, 'default_sample':
            max_intensity_def_sample.to_dict(orient='record'),
                             'outlier_samples': outlier_samples.to_dict(
                                 orient='record')})
    return outlier_test

